@@ -73,34 +73,7 @@ class CLIParser:
             help="the YellowDog Application secret",
             metavar="<app-secret>",
         )
-        parser.add_argument(
-            "--namespace",
-            "-n",
-            type=str,
-            required=False,
-            nargs="?",
-            const="",
-            help=(
-                "the namespace to use when creating and identifying entities;"
-                " this is set to '' if the option is provided without a value"
-            ),
-            metavar="<namespace>",
-        )
-        parser.add_argument(
-            "--tag",
-            "-t",
-            "--prefix",
-            type=str,
-            required=False,
-            nargs="?",
-            const="",
-            help=(
-                "the tag to use when naming, tagging, or selecting entities,"
-                " or the prefix (directory) when used with the object store;"
-                " this is set to '' if the option is provided without a value"
-            ),
-            metavar="<tag>",
-        )
+
         parser.add_argument(
             "--url",
             "-u",
@@ -149,6 +122,48 @@ class CLIParser:
 
         # Module-specific argument sets
 
+        if not any(
+            module in sys.argv[0]
+            for module in [
+                "boost",
+                "resize",
+                "create",
+                "remove",
+                "cloudwizard",
+                "follow",
+                "hold",
+                "start",
+            ]
+        ):
+            parser.add_argument(
+                "--namespace",
+                "-n",
+                type=str,
+                required=False,
+                nargs="?",
+                const="",
+                help=(
+                    "the namespace to use when creating and identifying entities;"
+                    " this is set to '' if the option is provided without a value"
+                ),
+                metavar="<namespace>",
+            )
+            parser.add_argument(
+                "--tag",
+                "-t",
+                "--prefix",
+                type=str,
+                required=False,
+                nargs="?",
+                const="",
+                help=(
+                    "the tag to use when naming, tagging, or selecting entities,"
+                    " or the prefix (directory) when used with the object store;"
+                    " this is set to '' if the option is provided without a value"
+                ),
+                metavar="<tag>",
+            )
+
         if any(module in sys.argv[0] for module in ["submit"]):
             parser.add_argument(
                 "--work-requirement",
@@ -322,6 +337,7 @@ class CLIParser:
                 "terminate",
                 "resize",
                 "cloudwizard",
+                "boost",
             ]
         ):
             parser.add_argument(
@@ -622,6 +638,20 @@ class CLIParser:
                 help="resize a compute requirement instead of a worker pool",
             )
 
+        if "boost" in sys.argv[0]:
+            parser.add_argument(
+                "allowance",
+                metavar="<allowance-ID>",
+                type=str,
+                help=("the YellowDog ID of the allowance to boost"),
+            )
+            parser.add_argument(
+                "boost_hours",
+                metavar="<boost hours>",
+                type=int,
+                help="the number of hours to boost the allowance by",
+            )
+
         if "shutdown" in sys.argv[0]:
             parser.add_argument(
                 "worker_pool_list",
@@ -1348,6 +1378,16 @@ class CLIParser:
     def object_path_pattern(self) -> Optional[str]:
         return self.args.pattern
 
+    @property
+    @allow_missing_attribute
+    def allowance(self) -> str:
+        return self.args.allowance
+
+    @property
+    @allow_missing_attribute
+    def boost_hours(self) -> int:
+        return self.args.boost_hours
+
 
 def lookup_module_description(module_name: str) -> Optional[str]:
     """
@@ -1392,6 +1432,8 @@ def lookup_module_description(module_name: str) -> Optional[str]:
         suffix = "starting held (paused) Work Requirements"
     elif "hold" in module_name:
         suffix = "holding (pausing) running Work Requirements"
+    elif "boost" in module_name:
+        suffix = "boosting Allowances"
 
     return None if suffix is None else prefix + suffix
 
