@@ -4,13 +4,21 @@
 A script to create or update YellowDog resources.
 """
 
-from typing import Dict, List
+from copy import deepcopy
+from datetime import datetime
+from typing import Dict, List, Optional
 
 import yellowdog_client.model as model
+from dateparser import parse as date_parse
 from requests import post
 from requests.exceptions import HTTPError
 from yellowdog_client.model import (
+    AccountAllowance,
     AddConfiguredWorkerPoolResponse,
+    AllowanceLimitEnforcement,
+    AllowanceResetType,
+    AwsFleetComputeSource,
+    AwsFleetPurchaseOption,
     CloudProvider,
     ImageOsType,
     Keyring,
@@ -18,33 +26,71 @@ from yellowdog_client.model import (
     MachineImageFamily,
     MachineImageGroup,
     NamespaceStorageConfiguration,
+    RequirementsAllowance,
+    SourceAllowance,
+    SourcesAllowance,
     WorkerPoolStatus,
     WorkerPoolSummary,
 )
 from yellowdog_client.model.exceptions import InvalidRequestException
 
 from yd_commands.interactive import confirmed
+from yd_commands.load_resources import load_resource_specifications
 from yd_commands.object_utilities import (
-    find_compute_source_id_by_name,
-    find_compute_template_id_by_name,
+    clear_compute_source_template_cache,
+    clear_image_family_search_cache,
+    find_compute_requirement_template_ids_by_name,
+    find_compute_source_template_id_by_name,
+    find_image_family_ids_by_name,
 )
 from yd_commands.printing import print_error, print_json, print_log, print_warning
-from yd_commands.resource_config import load_resource_specifications
+from yd_commands.settings import (
+    RN_ALLOWANCE,
+    RN_CONFIGURED_POOL,
+    RN_CREDENTIAL,
+    RN_IMAGE_FAMILY,
+    RN_KEYRING,
+    RN_REQUIREMENT_TEMPLATE,
+    RN_SOURCE_TEMPLATE,
+    RN_STORAGE_CONFIGURATION,
+)
 from yd_commands.wrapper import ARGS_PARSER, CLIENT, CONFIG_COMMON, main_wrapper
 
+CLEAR_CST_CACHE: bool = False  # Track whether the CST cache needs to be cleared
+CLEAR_IMAGE_FAMILY_CACHE: bool = False  # Track whether the IF cache needs to be cleared
+
 
 @main_wrapper
 def main():
-    resources = load_resource_specifications()
+    create_resources()
+
+
+def create_resources(
+    resources: Optional[List[Dict]] = None, show_secrets: bool = False
+):
+    """
+    Create a list of resources. Resources can be supplied as an argument, or
+    loaded from one or more files.
+    """
+    if resources is None:
+        resources = load_resource_specifications(creation_or_update=True)
+    else:
+        resources = deepcopy(resources)  # Avoid overwriting the input argument
+
     if ARGS_PARSER.dry_run:
         print_log(
-            "Dry run: displaying processed JSON resource specifications. Note:"
+            "Dry-run: displaying processed JSON resource specifications. Note:"
             " 'resource' property is removed."
         )
     for resource in resources:
         try:
             resource_type = resource.pop("resource")
-            if ARGS_PARSER.dry_run:
+            # There is potential additional processing for CRTs and CSTs
+            # Print JSON from within their functions
+            if ARGS_PARSER.dry_run and resource_type not in [
+                RN_REQUIREMENT_TEMPLATE,
+                RN_SOURCE_TEMPLATE,
+            ]:
                 print_json(resource)
                 continue
         except KeyError:
@@ -53,27 +99,29 @@ def main():
                 f" specification: {resource}"
             )
             continue
-        if resource_type == "ComputeSourceTemplate":
-            create_compute_source(resource)
-        elif resource_type == "ComputeRequirementTemplate":
-            create_cr_template(resource)
-        elif resource_type == "Keyring":
-            create_keyring(resource)
-        elif resource_type == "Credential":
+        if resource_type == RN_SOURCE_TEMPLATE:
+            create_compute_source_template(resource)
+        elif resource_type == RN_REQUIREMENT_TEMPLATE:
+            create_compute_requirement_template(resource)
+        elif resource_type == RN_KEYRING:
+            create_keyring(resource, show_secrets)
+        elif resource_type == RN_CREDENTIAL:
             create_credential(resource)
-        elif resource_type == "MachineImageFamily":
+        elif resource_type == RN_IMAGE_FAMILY:
             create_image_family(resource)
-        elif resource_type == "NamespaceStorageConfiguration":
+        elif resource_type == RN_STORAGE_CONFIGURATION:
             create_namespace_configuration(resource)
-        elif resource_type == "ConfiguredWorkerPool":
+        elif resource_type == RN_CONFIGURED_POOL:
             create_configured_worker_pool(resource)
+        elif resource_type == RN_ALLOWANCE:
+            create_allowance(resource)
         else:
             print_error(f"Unknown resource type '{resource_type}'")
 
 
-def create_compute_source(resource: Dict):
+def create_compute_source_template(resource: Dict):
     """
-    Create or update a Compute Source using a resource specification.
+    Create or update a Compute Source Template using a resource specification.
     Handles all Source types.
     """
     try:
@@ -83,6 +131,28 @@ def create_compute_source(resource: Dict):
     except KeyError as e:
         raise Exception(f"Expected property to be defined ({e})")
 
+    # Allow image families to be referenced by name rather than ID
+    global CLEAR_IMAGE_FAMILY_CACHE
+    if CLEAR_IMAGE_FAMILY_CACHE:  # Update the IF cache if required
+        clear_image_family_search_cache()
+        CLEAR_IMAGE_FAMILY_CACHE = False
+    image_id = source.get("imageId")
+    if image_id is not None:
+        if "ydid:imgfam:" not in image_id:
+            image_family_ids = find_image_family_ids_by_name(
+                client=CLIENT, image_family_name=image_id
+            )
+            if len(image_family_ids) > 0:
+                source["imageId"] = image_family_ids[0]
+                print_log(
+                    f"Replaced imageId name '{image_id}' with ID {image_family_ids[0]}"
+                )
+
+    if ARGS_PARSER.dry_run:
+        resource["source"] = source
+        print_json(resource)
+        return
+
     # Create the Compute Source
     compute_source = get_model_object(source_type, source)
 
@@ -92,31 +162,35 @@ def create_compute_source(resource: Dict):
     )
 
     # Check for an existing ID
-    source_id = find_compute_source_id_by_name(CLIENT, name)
+    source_id = find_compute_source_template_id_by_name(CLIENT, name)
     if source_id is None:
         compute_source = CLIENT.compute_client.add_compute_source_template(
             compute_source_template
         )
         print_log(
-            f"Created Compute Source '{compute_source.source.name}'"
+            f"Created Compute Source Template '{compute_source.source.name}'"
             f" ({compute_source.id})"
         )
     else:
-        if not confirmed(f"Update existing Source Template '{name}'?"):
+        if not confirmed(f"Update existing Compute Source Template '{name}'?"):
             return
         compute_source_template.id = source_id
         compute_source = CLIENT.compute_client.update_compute_source_template(
             compute_source_template
         )
         print_log(
-            f"Updated existing Compute Source '{compute_source.source.name}'"
+            f"Updated existing Compute Source Template '{compute_source.source.name}'"
             f" ({compute_source.id})"
         )
-    if ARGS_PARSER.quiet:
+
+    global CLEAR_CST_CACHE
+    CLEAR_CST_CACHE = True
+
+    if ARGS_PARSER.quiet and compute_source.id is not None:
         print(compute_source.id)
 
 
-def create_cr_template(resource: Dict):
+def create_compute_requirement_template(resource: Dict):
     """
     Create or update a Compute Requirement Template. Handles all
     Compute Requirement types.
@@ -127,33 +201,93 @@ def create_cr_template(resource: Dict):
     except KeyError as e:
         raise Exception(f"Expected property to be defined ({e})")
 
+    # Allow source templates to be referenced by name instead of ID:
+    # substitute ID for name
+    global CLEAR_CST_CACHE
+    if CLEAR_CST_CACHE:  # Update the CST cache if required
+        clear_compute_source_template_cache()
+        CLEAR_CST_CACHE = False
+
+    # Dynamic templates don't have 'sources'; return '[]'
+    counter = 0
+    for source in resource.get("sources", []):
+        template_name_or_id = source["sourceTemplateId"]
+        if "ydid:cst:" not in template_name_or_id:
+            template_id = find_compute_source_template_id_by_name(
+                client=CLIENT, name=template_name_or_id
+            )
+            if template_id is None:
+                print_error(
+                    f"Compute Source Template name '{template_name_or_id}' not found"
+                )
+                return
+            source["sourceTemplateId"] = template_id
+            counter += 1
+
+    if counter > 0:
+        print_log(f"Replaced {counter} Compute Source Template name(s) with ID(s)")
+
+    # Allow image families to be referenced by name rather than ID
+    global CLEAR_IMAGE_FAMILY_CACHE
+    if CLEAR_IMAGE_FAMILY_CACHE:  # Update the IF cache if required
+        clear_compute_source_template_cache()
+        CLEAR_IMAGE_FAMILY_CACHE = False
+
+    images_id = resource.get("imagesId")
+    if images_id is not None:
+        if "ydid:imgfam:" not in images_id:
+            image_family_ids = find_image_family_ids_by_name(
+                client=CLIENT, image_family_name=images_id
+            )
+            if len(image_family_ids) > 0:
+                resource["imagesId"] = image_family_ids[0]
+                print_log(
+                    f"Replaced imagesId name '{images_id}' with ID"
+                    f" {image_family_ids[0]}"
+                )
+
+    if ARGS_PARSER.dry_run:
+        print_json(resource)
+        return
+
     compute_template = get_model_object(type, resource)
 
     # Check for an existing ID
-    template_id = find_compute_template_id_by_name(CLIENT, name)
-    if template_id is None:
+    template_ids = find_compute_requirement_template_ids_by_name(CLIENT, name)
+    if len(template_ids) == 0:
         template = CLIENT.compute_client.add_compute_requirement_template(
             compute_template
         )
         print_log(
             f"Created Compute Requirement Template '{template.name}' ({template.id})"
         )
+        if ARGS_PARSER.quiet:
+            print(template.id)
     else:
-        compute_template.id = template_id
-        if not confirmed(f"Update existing Compute Requirement Template '{name}'?"):
-            return
-        template = CLIENT.compute_client.update_compute_requirement_template(
-            compute_template
-        )
-        print_log(
-            f"Updated existing Compute Requirement Template '{template.name}'"
-            f" ({template.id})"
-        )
-    if ARGS_PARSER.quiet:
-        print(template.id)
+        if len(template_ids) > 1:
+            print_warning(
+                f"{len(template_ids)} Compute Requirement Templates with the name"
+                f" '{name}'"
+            )
+        for template_id in template_ids:
+            compute_template.id = template_id
+            if not confirmed(
+                f"Update existing Compute Requirement Template '{name}'"
+                f" ({template_id})?"
+            ):
+                return
+            template = CLIENT.compute_client.update_compute_requirement_template(
+                compute_template
+            )
+            print_log(
+                f"Updated existing Compute Requirement Template '{template.name}'"
+                f" ({template.id})"
+            )
+            if ARGS_PARSER.quiet:
+                print(template.id)
 
 
-def create_keyring(resource: Dict):
+def create_keyring(resource: Dict, show_secrets: bool = False):
     """
     Create or delete/recreate a Keyring.
     """
@@ -173,17 +307,16 @@ def create_keyring(resource: Dict):
 
     try:
         keyring, keyring_password = create_keyring_via_api(name, description)
-        if ARGS_PARSER.show_keyring_passwords:
-            if not ARGS_PARSER.quiet:
-                print_log(
-                    f"Created Keyring '{name}' ({keyring.id}): Password ="
-                    f" {keyring_password}"
-                )
-            else:
-                print(f"{keyring.id} : Password = {keyring_password}")
-        print_log(f"Created Keyring '{name}'")
+        keyring_password = (
+            keyring_password
+            if ARGS_PARSER.show_keyring_passwords or show_secrets
+            else "<REDACTED>"
+        )
+        print_log(
+            f"Created Keyring '{name}' ({keyring.id}): Password = {keyring_password}"
+        )
         if ARGS_PARSER.quiet:
-            print(keyring.id)
+            print(f"{keyring.id} {keyring_password}")
     except Exception as e:
         print_error(f"Failed to create Keyring '{name}': {e}")
 
@@ -225,6 +358,7 @@ def create_credential(resource: Dict):
         CLIENT.keyring_client.put_credential_by_name(keyring_name, credential)
         print_log(f"Added Credential '{name}' to Keyring '{keyring_name}'")
     except HTTPError as e:
+        print_error(f"Failed to add Credential '{name}' to Keyring '{keyring_name}'")
         if e.response.status_code == 400:
             print_error(f"{e.response.text}")
         elif e.response.status_code == 404:
@@ -240,10 +374,18 @@ def create_image_family(resource):
     try:
         family_name = resource["name"]
         namespace = resource["namespace"]
-        os_type = ImageOsType[resource.pop("osType")]  # Change to Enum
+        os_type_str = resource.pop("osType")
     except KeyError as e:
         raise Exception(f"Expected property to be defined ({e})")
 
+    try:
+        os_type = ImageOsType[os_type_str]  # Change to Enum
+    except KeyError:
+        raise Exception(
+            f"Property 'osType' has invalid value '{os_type_str}'; valid values are"
+            f" {[e.value for e in ImageOsType]}"
+        )
+
     # Start by updating the outer Image Family
     image_family = get_model_object("MachineImageFamily", resource, osType=os_type)
 
@@ -294,6 +436,9 @@ def create_image_family(resource):
         image_group.osType = ImageOsType[str(image_group.osType)]  # Replace with Enum
         _create_image_group(namespace, image_family, image_group)
 
+    global CLEAR_IMAGE_FAMILY_CACHE
+    CLEAR_IMAGE_FAMILY_CACHE = True
+
 
 def _create_image_group(
     namespace: str, image_family: MachineImageFamily, image_group: MachineImageGroup
@@ -449,20 +594,133 @@ def create_configured_worker_pool(resource: Dict):
         print_error(f"Unable to created Configured Worker Pool '{name}'")
 
 
+def create_allowance(resource: Dict):
+    """
+    Create an allowance.
+    """
+    try:
+        type = resource.pop("type").split(".")[-1]  # Extract type
+    except KeyError as e:
+        raise Exception(f"Expected property to be defined ({e})")
+
+    if type == "SourcesAllowance":
+        template_name_or_id = resource.get("sourceCreatedFromId", None)
+        if template_name_or_id is not None:
+            if "ydid:cst:" not in template_name_or_id:
+                template_id = find_compute_source_template_id_by_name(
+                    client=CLIENT, name=template_name_or_id
+                )
+                if template_id is None:
+                    print_error(
+                        f"Compute Source Template name '{template_name_or_id}' not found"
+                    )
+                    return
+                print_log(
+                    f"Replaced Source Template name '{template_name_or_id}'"
+                    f" with ID {template_id}"
+                )
+                resource["sourceCreatedFromId"] = template_id
+
+    elif type == "RequirementsAllowance":
+        template_name_or_id = resource.get("requirementCreatedFromId", None)
+        if template_name_or_id is not None:
+            if "ydid:crt:" not in template_name_or_id:
+                template_ids = find_compute_requirement_template_ids_by_name(
+                    client=CLIENT, name=template_name_or_id
+                )
+                if len(template_ids) == 0:
+                    print_error(
+                        f"Compute Requirement Template name '{template_name_or_id}' not found"
+                    )
+                    return
+                template_id = template_ids[0]
+                print_log(
+                    f"Replaced Requirement Template name '{template_name_or_id}'"
+                    f" with ID {template_id}"
+                )
+                resource["requirementCreatedFromId"] = template_id
+
+    allowance = CLIENT.allowances_client.add_allowance(get_model_object(type, resource))
+
+    print_log(f"Created allowance ID {allowance.id}")
+
+    if ARGS_PARSER.quiet and allowance.id is not None:
+        print(allowance.id)
+
+
 def get_model_object(classname: str, resource: Dict, **kwargs):
     """
     Return a populated YellowDog model object. Handle unexpected keywords.
     """
+
+    def _fix_aws_fleet_properties():
+        if isinstance(model_object, AwsFleetComputeSource):
+            try:
+                model_object.purchaseOption = AwsFleetPurchaseOption[
+                    str(model_object.purchaseOption)
+                ]
+            except KeyError:
+                raise Exception(
+                    "Invalid AWS Fleet Compute Source Purchase Option property: "
+                    f"'{str(model_object.purchaseOption)}'"
+                )
+
+    def _fix_allowance_properties():
+        if (
+            isinstance(model_object, SourceAllowance)
+            or isinstance(model_object, SourcesAllowance)
+            or isinstance(model_object, RequirementsAllowance)
+            or isinstance(model_object, AccountAllowance)
+        ):
+            try:
+                effective_from = resource.get("effectiveFrom", None)
+                if effective_from is not None:
+                    model_object.effectiveFrom = date_parse(effective_from)
+                    if model_object.effectiveFrom is None:
+                        raise Exception(
+                            f"Unable to parse 'effectiveFrom' date '{effective_from}'"
+                        )
+                    print_log(
+                        f"Property 'effectiveFrom' = '{effective_from}' set to "
+                        f"'{str(model_object.effectiveFrom)}'"
+                    )
+                effective_until = resource.get("effectiveUntil", None)
+                if effective_until is not None:
+                    model_object.effectiveUntil = date_parse(effective_until)
+                    if model_object.effectiveUntil is None:
+                        raise Exception(
+                            f"Unable to parse 'effectiveUntil' date '{effective_until}'"
+                        )
+                    print_log(
+                        f"Property 'effectiveUntil' = '{effective_until}' set to "
+                        f"'{str(model_object.effectiveUntil)}'"
+                    )
+                model_object.limitEnforcement = AllowanceLimitEnforcement(
+                    model_object.limitEnforcement
+                )
+                model_object.resetType = AllowanceResetType(model_object.resetType)
+            except KeyError as e:
+                raise Exception(f"Invalid property: {e}")
+
     while True:
         try:
-            object = get_model_class(classname)(**resource, **kwargs)
-            return object
+            model_object = get_model_class(classname)(**resource, **kwargs)
+            _fix_aws_fleet_properties()
+            _fix_allowance_properties()
+            return model_object
         except Exception as e:
-            # Unexpected keyword argument Exception of form:
-            # __init__() got an unexpected keyword argument 'keyword'
-            keyword = str(e).split("'")[1]
-            print_warning(f"Ignoring unexpected property '{keyword}'")
-            resource.pop(keyword)
+            # Unexpected/missing keyword argument Exception of form:
+            # __init__() got an unexpected keyword argument 'keyword', or
+            # __init__() missing 1 required positional argument: 'credential'
+            if "unexpected" in str(e):
+                keyword = str(e).split("'")[1]
+                print_warning(f"Ignoring unexpected property '{keyword}'")
+                resource.pop(keyword)
+            elif "missing" in str(e):
+                keyword = str(e).split("'")[1]
+                raise Exception(f"Missing expected property '{keyword}'")
+            else:
+                raise e
 
 
 def get_model_class(classname: str):
